# 16-08-04
#v1.0.4

# filestruct
# Classes and functions to represent and compare file structures.
# Also a basic markup language to save them and record changes.
# FSDM - file structure description markup

# This could be the basis of, for example -
# a remote directory sync, incremental archive, version control system etc
# In fact anything where changes to a file system need to be monitored.

# Also a simple command line utility for doing 'remote file sync', by monitoring a directory structure for changes
# and only saving/implementing the changes.

# Copyright Michael Foord
# You are free to modify, use and relicense this code.
# No warranty express or implied for the accuracy, fitness to purpose or otherwise for this code....
# Use at your own risk !!!

# If you have any bug reports, questions or suggestions please contact me.
# If you would like to be notified of bugfixes/updates then please contact me and I'll add you to my mailing list.
# E-mail michael AT foord DOT me DOT uk
# Maintained at www.voidspace.org.uk/atlantibots/pythonutils.html

"""
filestruct is a library of functions for describing file structures.
It defines a basic markup language - FSDM - file structure description markup.
We can create an object representing a file structure (DirObj) and write out a representation of this structure using FSDM.

We can then read this FSDM back into recreate the object. (Using ReadFSDM which is a subclass of DirObj).

We can also compare two DirObjs using a ChangeObject.
A ChangeObject can describe the changes using a similar markup - it writes an FSCO - file structure change object.
It can also create a zip file with all the additonal files in the newer DirObj.
From this zip file and an FSCO script it can implement all the changes (including file and directory deletes) on another file structure - which is used when filestruct is run from teh command line to remotely sync two directories.

Speed of indexing large structures is very dependent on disk-access times.
Indexing an 873MByte directory, with 16647 files in 668 folders took 31.25 seconds.
With psyco enabled it only took 20.14 seconds.
A repeat of the read (most of the information cached) took 7 seconds.
Parsing the FSDM structure representing the same File Structure took only 2.125 seconds.

See the docs for a full description of the two markups and the methods/attributes of DirObj and ChangeObject instances.

The objects here are subclasses of strings - so they need Python 2.2 +
"""

# MOD_CONST
# If PSYCOON is set to 0 then we won't try and import the psyco module
# IF importing fails, PSYCOON is set to 0
PSYCOON = 1
if PSYCOON:
    try:
        import psyco
        psyco.full()
    except:
        PSYCOON = 0

import sys, os, time
from os.path import join, islink, basename, isfile, isdir, getmtime, getctime, getsize, abspath, dirname, exists
from shutil import rmtree

try:
    from ordereddict import oDict
    dictmeth = oDict
except:
    dictmeth = dict

class DirObj(str):
    """Produce an object that efficiently represents a directory and the files in it.
    Will ignore symbolic links.
    Each subdirectory will itself be a directory object.
    If this is the root directory of the archive (relpath='') you can pass in the keyword 'name' which is the archive name.
    """
    def __new__(self, directorypath, relpath='', parent=None, incabs=0, **keywargs):
        """This method overloads the string __new__ method and stops our extra arguments confusing it."""
        if keywargs.has_key('abs') or incabs:           # the ReadFSDM doesn't have this option.
            thepath = abspath(join(directorypath, relpath))
        else:
            thepath = relpath
        return str.__new__(self, thepath)
        
    def __init__(self, directorypath, relpath='', parent=None, incabs=0, **keywargs):
        thisdir = join(directorypath, relpath)
        if not isdir(thisdir): raise TypeError(': Can only create a DirObj with a valid directory - %s' % directorypath)
        self.files = dictmeth()
        self.dirs = dictmeth()
        self.size = 0
        self.name = basename(thisdir)
        self.path = relpath
        self.parent = parent                    # link to the parent DirObj - or None if this is the root directory (the DirObj structure is relative)
        if keywargs.has_key('abs'): incabs = 1
        self.numfiles = 0
        self.numdirs = 0

        for apath in os.listdir(thisdir):
            thepath = join(thisdir, apath)
            if islink(thepath): continue
            if isdir(thepath):
                thedir = DirObj(directorypath, join(relpath, apath), self, incabs)
                self.dirs[apath] = thedir
                self.size += thedir.size
                self.numdirs += (thedir.numdirs + 1)
                self.numfiles += thedir.numfiles
            else:
                thisfile = self.makefiledict(apath, thepath, relpath)
                self.files[apath] = thisfile
                self.size += thisfile['size']
                self.numfiles += 1
            
        self.ctime = getctime(thisdir)
        self.mtime = getmtime(thisdir)
        if relpath == '':                   # set the archive name for the rootpath
            self.archivename = keywargs.get('name','')
            self.creationtime = time.time()
            self.name = ''
            self.abspath = abspath(directorypath)              
            
        self.indent = 0         # set this attribute to -1 before calling the write method if you want no indentation in the out structure

    def makefiledict(self, filename, filepath, relpath):
        """Form a dictionary with the basic attributes of a file."""
        filedict = dictmeth()
        filedict['path'] = relpath
        filedict['name'] = filename
        filedict['size'] = getsize(filepath)
        filedict['ctime'] = getctime(filepath)
        filedict['mtime'] = getmtime(filepath)
        return filedict

    def writefile(self, filedict, indent=None):
        """Write out the attributes representing a file.
        Uses the FSDM markup system.
        Observes the indent level - 4 spaces per level. A value of -1 means no indentation."""
        if indent == None: indent = self.indent
        if indent != -1:
            indent = indent * '    '
        else:
            indent = ''
        outstuff = [maketag('file')]
        outstuff.append(maketag('name', filedict['name']))
        outstuff.append(maketag('size', filedict['size']))
        outstuff.append(maketag('ctime', filedict['ctime']))
        outstuff.append(maketag('mtime', filedict['mtime']))
        outstuff.append(maketag('/file'))        
        return [indent + line for line in outstuff]

    def write(self, indent=None):
        """Write out the attributes representing the DirObj and any nested files/directories.
        Uses the FSDM markup system.
        Observes the indent level - 4 spaces per level. A value of -1 means no indentation.
        If the path attribtue of the directory object is '' then it is the root directory.
        This means that write will put the start and end tags in place and not the directory tag."""
        outstuff = []
        if indent == None: indent = self.indent
        if indent != -1:
            theindent = indent * '    '
        else:
            theindent = ''
        
        if not self.path: # this is the root directory
##            print len(self.dirs)
##            print len(self.files)
##            for entry in  self.files.items(): print entry
            outstuff.append(theindent+maketag('fsdm'))
            outstuff.append(theindent+maketag('aname', self.archivename))
            outstuff.append(theindent+maketag('time', self.creationtime))
            outstuff.append(theindent+maketag('date', time.ctime(self.creationtime)))               # included for readability, never actually read back ! time and date are identical
            outstuff.append(theindent+maketag('abspath', self.abspath))
        else:                                   
            outstuff.append(theindent+maketag('directory'))

        outstuff.append(theindent+maketag('name', self.name))
        outstuff.append(theindent+maketag('size', self.size))
        outstuff.append(theindent+maketag('ctime', self.ctime))
        outstuff.append(theindent+maketag('mtime', self.mtime))
        outstuff.append(theindent+maketag('numfiles', self.numfiles))
        outstuff.append(theindent+maketag('numdirs', self.numdirs))
        if indent != -1:
            indent += 1
            theindent = indent * '    '
        for entry in self.files:
            outstuff += self.writefile(self.files[entry], indent)          # each file is basically a dictionary of attributes
        for entry in self.dirs:
            outstuff += self.dirs[entry].write(indent)

        if indent != -1:
            indent -= 1
            theindent = indent * '    '

        if self.path:
            outstuff.append(theindent+maketag('/directory'))
        else:                                   # this is the root directory
            outstuff.append(theindent+maketag('/fsdm'))
                    
        return outstuff


class ReadFSDM(DirObj):
    """Read and parse a saved file structure in the FSDM format.
    It takes the FSDM structure as a list of lines and the relative path to it.
    The returned object is effectively the DirObj it was written from.
    It assumes a valid structure - and will handle bad structures badly.
    """
    def __new__(self, indir, relpath='', parent=None, donestart=0):
        return str.__new__(self, relpath)
    
    def __init__(self, indir, relpath='', parent=None, donestart=0):        
        self.files = dictmeth()
        self.dirs = dictmeth()
        self.parent = parent
        self.path = relpath
        self.indent = 0

        self.instruct = indir = [line.strip() for line in indir]            # remove indentation and empty lines
        self.index = -1
        while self.index < len(indir)-1:          # indir is the FSDM structure we are parsing - it will either be just a direcotry structure or a whole FSDM struct
            self.index += 1
            line = indir[self.index]
            tag, value = handletag(line)
            if tag ==  'fsdm' and not donestart:
                donestart = 1
                continue
            elif not donestart:
                continue
            elif not tag:               # invalid tag or line - raise exception ?
                continue
            elif tag=='file':
                filechunk = self._findchunk('file')
                if filechunk:                            # if not filechunk then the structure is broken - raise exception ?
                    filedict = self.makefiledict(filechunk, relpath)
                    self.files[filedict['name']] = filedict
                continue
            elif tag=='directory':
                dirchunk = self._findchunk('directory')
                if dirchunk:                            # if not dirchunk then the structure is broken - raise exception ?
                    dirchunk, thename = dirchunk        # we use the look ahead in _findchunk to read the directory name, this is needed to set the path
                    adir = ReadFSDM(dirchunk, join(relpath, thename), self, 1)
                    self.dirs[thename] = adir
                continue
           # the above chunks alter self.index appropriately
            elif tag == 'aname':
                self.archivename = value
                continue
            elif tag == 'name':
                self.name = value
                continue
            elif tag == 'size':
                self.size = int(value)
                continue
            elif tag == 'numfiles':
                self.numfiles = int(value)
                continue
            elif tag == 'numdirs':
                self.numdirs = int(value)
                continue
            elif tag == 'ctime':
                self.ctime = int(value)
                continue
            elif tag == 'mtime':
                self.mtime = int(value)
                continue
            elif tag == 'abspath':
                self.abspath = value
                continue
            elif tag =='time':
                self.creationtime = float(value)
                continue
            elif tag == 'date':                     # date is a tag for readability only
                continue
            elif tag == '/fsdm':                    # end of structure
                break
            

    def _findchunk(self, param):
        """Given a lowercase param to search for, it will find the chunk with the closing param tag.
        Allows for nested tags of the same type. (Which you don't want for file but never mind).
        This now features a lookahead for directory names.....
        """
        lookahead = 0
        thename = ''
        closeparam = '/'+param
        numtag = 1
        outstuff = []
        index = self.index
        while self.index < len(self.instruct) - 1:
            self.index += 1
            line = self.instruct[self.index]
            tag, value = handletag(line)
            if tag == 'name' and not lookahead:         # the first name we come to will be the directory name
                lookahead = 1
                thename = value
            if not tag:                         # invalid line - oh dear
                outstuff.append(line)
                continue                    # invalid tag
            if tag == param:
                outstuff.append(line)
                numtag += 1
                continue
            elif tag == closeparam:
                outstuff.append(line)
                numtag -= 1
                if numtag == 0:
                    if param == 'directory':
                        return (outstuff, thename)
                    return outstuff
                continue
            else:
               outstuff.append(line) 

        self.index = index   # if get this far, then the structure is invalid - not enough closing parameters
        return None

    def makefiledict(self, infile, relpath):
        """Make a file dict from the attributes in the file chunk."""
        filedict = dictmeth()
        filedict['path'] = relpath
        index = -1
        while index < len(infile)-1:          # infile is the FSDM structure we are parsing, should just be one file chunk here 
            index += 1
            line = infile[index]
            tag, value = handletag(line)
            if not tag: continue                    # invalid tag or line
            if tag == 'name':
                filedict['name'] = value
            elif tag == 'size':
                filedict['size'] = int(value)
            elif tag == 'ctime':
                filedict['ctime'] = int(value)
            elif tag == 'mtime':
                filedict['mtime'] = int(value)
        return filedict


class ChangeObject:
    """For comparing directory objects.
    Methods to generate both a markup script (FSCO) describing the changes 
    and a zip file containing the new files.

    Also to read a saved FSCO script (File Structure Change Object) and accompanying zip file and implement the changes.
    """
    def __init__(self):
        self.newfiles = []        
        self.delfiles = []
        self.modfiles = []
        self.newdirs = []
        self.deldirs = []
        self.warnings = []
        self.archivename = ''
        self.debug = 0

    def compare(self, dirobj1, dirobj2):
        """Compare two DirObj s and record the changes.
        dirobj2 must be younger (newer) than dirobj1 or ValueError will be raised.
        This is to help prevent disastrous mistakes being made.
        This process *may* be destructive on the DirObj.....
        """
        if not hasattr(dirobj1, 'creationtime'):
            raise TypeError(": dirobj1 doesn't appear to be a valid DirObj.")
        if not hasattr(dirobj2, 'creationtime'):
            raise TypeError(": dirobj2 doesn't appear to be a valid DirObj.")
        if dirobj1.creationtime >= dirobj2.creationtime:
            raise ValueError(": dirobj2 must be newer than dirobj1.")

        self.newfiles = []        
        self.delfiles = []
        self.modfiles = []
        self.newdirs = []
        self.deldirs = []
        self.warnings = []
        self.archivename = dirobj2.archivename
        self.__comparedirlist(dirobj1, dirobj2)

    def __comparefilelist(self, dirobj1, dirobj2):
        """Compare file lists from two directory objects.
        They should both represent the same directory.
        """
        relpath = dirobj2.path        # dirobj1 and dirobj2 path and name attributes should be the same !
        fdict1 = dirobj1.files 
        fdict2 = dirobj2.files 

        for entry in fdict1:
            if entry not in fdict2:
                self.delfiles.append(join(relpath, entry))
            else:
                if dirobj1.files[entry] != dirobj2.files[entry]:        # this assumes *every* entry in the filedict will be the same				# if the two file dicts aren't the same... then the file has changed
                    self.modfiles.append(join(relpath, entry))
                del fdict2[entry]

        for entry in fdict2:             # these are the ones left
            self.newfiles.append(join(relpath, entry))


    def __comparedirlist(self, dirobj1, dirobj2):
        """Compare the directory lists of a DirObj. 
        Each directory in turn may need to call _comparedirlist.
        This fills in :
        self.newfiles = []        
        self.delfiles = []
        self.modfiles = []
        self.newdirs = []
        self.deldirs = []
        """
        ddict1 = dirobj1.dirs
        ddict2 = dirobj2.dirs
        self.__comparefilelist(dirobj1, dirobj2)  # compare the files in this directory
        for entry in ddict1:
            if entry not in ddict2:
                self.deldirs.append(ddict1[entry].path)
            else:
                self.__comparedirlist(ddict1[entry], ddict2[entry])
                del ddict2[entry]

        for entry in ddict2:             # these are the ones left
            self.newdirs.append(ddict2[entry].path)

    def write(self):
        """Write a list of changes between dirobj1 and dirobj2.
        It returns None if there are no changes.
        (Unless self.debug is set to 0)
        """
        if not self.debug and not self.newfiles and not self.delfiles and not self.modfiles and not self.newdirs and not self.deldirs:
            return None
        outstuff = []
        outstuff.append(maketag('fsco'))         # file structure change object
        outstuff.append(maketag('aname', self.archivename))  
        outstuff.append(maketag('time', time.time())) 
        outstuff.append(maketag('date', time.ctime())) 
        outstuff.append(maketag('newfiles'))  
        for entry in self.newfiles:       
            outstuff.append(entry)
        outstuff.append(maketag('/newfiles'))  
        outstuff.append(maketag('delfiles'))  
        for entry in self.delfiles:       
            outstuff.append(entry)
        outstuff.append(maketag('/delfiles'))  
        outstuff.append(maketag('modfiles'))  
        for entry in self.modfiles:       
            outstuff.append(entry)
        outstuff.append(maketag('/modfiles'))  
        outstuff.append(maketag('newdirs'))  
        for entry in self.newdirs:       
            outstuff.append(entry)
        outstuff.append(maketag('/newdirs'))  
        outstuff.append(maketag('deldirs'))  
        for entry in self.deldirs:       
            outstuff.append(entry)
        outstuff.append(maketag('/deldirs'))  

        outstuff.append(maketag('/fsco'))  
        return outstuff

    def read(self, infile):
        """Read in an FSCO file and rebuild a change object.
        This can then call the makechange method.
        """
        self.newfiles = []        
        self.delfiles = []
        self.modfiles = []
        self.newdirs = []
        self.deldirs = []
        self.warnings = []
        self.instruct = infile = [line.strip() for line in infile]
        self.index = -1
        while self.index < len(infile)-1:
            self.index += 1
            line = infile[self.index]
            tag, value = handletag(line)
            if not tag: continue                    # error - throw exception ?
            if tag == 'fsco':
                continue
            if tag.startswith('/'):                 # closing tags are irrelevant in the FSCO structure.... so far anyway (no nesting)
                if tag == '/fsco': break
                continue
            elif tag == 'newfiles':
                self.newfiles = self._findchunk()
            elif tag == 'delfiles':
                self.delfiles = self._findchunk()
            elif tag == 'modfiles':
                self.modfiles = self._findchunk()
            elif tag == 'newdirs':
                self.newdirs = self._findchunk()
            elif tag == 'deldirs':
                self.deldirs = self._findchunk()
            elif tag == 'aname':
                self.archivename = value
            elif tag == 'time': continue            # we ignore the following tags
            elif tag == 'date': continue

    def _findchunk(self):
        """For the ChangeObject - this just reads the file into a list until it finds another tag.
        """
        outstuff = []
        while self.index < len(self.instruct) - 1:
            self.index += 1
            line = self.instruct[self.index]
            if not line.startswith('<'):
                outstuff.append(line)
            else:
                break
        return outstuff

    def buildzip(self, directorypath, filename, warnings=1):
        """From a change object it will attempt to build a zip file with new and changed files.
        It needs an absolute path to the file structure (all file structures are relative),
        a filename to save the zip as - and whether you want warnings printed (default is on).
        If you set warnings to 2 it will raise exceptions for serious errors.
        """
        from zipfile import ZipFile, ZipInfo,ZIP_DEFLATED
        
        zip = ZipFile(filename, 'w',ZIP_DEFLATED)
        currentdir = os.getcwd()
        os.chdir(directorypath)         # so that file access is relative
        
        for entry in  self.newfiles + self.modfiles:
            try:
                zip.write(entry) 
            except:
                error = 'Error archiving file %s.' % entry
                if warnings == 1:
                    self.warnings.append(error)
                elif warnings == 2:
                    print error
                    raise

        for entry in self.newdirs:
            for thisfile in walkfiles(entry):            # this omits empty directories
                try:
                    zip.write(thisfile) 
                except:
                    error = 'Error archiving file %s.' % thisfile
                    if warnings == 1:
                        self.warnings.append(error)
                    elif warnings == 2:
                        print error
                        raise
                    
            for emptydir in walkemptydirs(entry):       # this yields empty directories
                try:
                    zi = ZipInfo(emptydir+'/')          # and this adds them to the zipfile
                    zip.writestr(zi,'')
                except:
                    error = 'Error archiving directory %s.' % emptydir
                    if warnings == 1:
                        self.warnings.append(error)
                    elif warnings == 2:
                        print error
                        raise            
                
        zip.close()
        os.chdir(currentdir)

    def build(self, directorypath, filename, warnings=1):
        """A combination of the buildzip and write methods.
        It combines an FSCO script and a zip of all the files necessary to implement the changes - as a single zip.
        Can be used by the rebuild method.
        If there are no changes (i.e. zipfile not made) it reurns False, else it returns True."""
        self.warnings = []
        from zipfile import ZipFile, ZipInfo, ZIP_DEFLATED
        
        script = self.write()
        if not script: return False
        scriptfile = 'FSCO.txt'
        temp1 = 0
        if exists(scriptfile):
            temp1 = readb(scriptfile)
        writel(scriptfile, script)

        archive = 'filearchive.zip'
        temp2 = 0
        if exists(archive):
            temp2 = readb(archive)
        self.buildzip(directorypath, archive, warnings)
        zip = ZipFile(filename, 'w', ZIP_DEFLATED)
        zip.write(scriptfile)
        zip.write(archive)
        zip.close()
        os.remove(scriptfile)
        os.remove(archive)
        if temp1: writeb(scriptfile, temp1)
        if temp2: writeb(archive, temp2)
        return True

    def rebuild(self, directorypath, filename, warnings=1):
        """Given a zip file created by the build method - which combines an FSCO structure and the files necessary -
        it reads the FSCO structure and implements the changes.
        WARNING - will delete files and directories if instructed to."""
        self.warnings = []
        from zipfile import ZipFile, ZipInfo, ZIP_DEFLATED

        tempdir = self.maketempdir()
        if tempdir == None:
            tempdir = ''
            if warnings == 1:
                self.warnings.append('Error creating temporary directory.')
            self.inzip = 1
        zip = ZipFile(filename, 'r')
        filelist = zip.namelist()
        if 'filearchive.zip' not in filelist or 'FSCO.txt' not in filelist:
            self.warnings.append("%s isn't an FSCO archive." % filename)

        arcfile = join(tempdir, 'filearchive.zip')
        fscoscript = zip.read('FSCO.txt').splitlines()
        writeb(arcfile, zip.read('filearchive.zip'))
        zip.close()
        self.read(fscoscript)
        self.makechange(directorypath, arcfile, warnings)
        os.remove(arcfile)
        if tempdir: os.rmdir(tempdir)

    def maketempdir(self):
        """Creates a temporary directory in the current directory.
        Used for unzipping files into.
        Creates the directory and returns the filename as a path object (or None).
        XX Will fail on a CD drive. Other standard library methods are deprecated or return an open file rather than a flename. XX
        """
        dirname = 'mjf-temp'
        number = -1
        thisdir = dirname
        while exists(thisdir):
            number += 1
            thisdir = dirname + str(number)
        try:
            os.mkdir(thisdir)
        except:
            return None
        else:
            return thisdir

    def makechange(self, directorypath, filename, warnings=1):
        """makechange takes a zip file and implements the changes in the change object.
        It is expected that the changes will have already been read from an FSCO structure.
        WARNING - will delete files and directories if instructed to.
        Doesn't reset self.warnings as may be called from rebuild which can make entries in self.warnings.
        """
        from zipfile import ZipFile, ZipInfo
        
        zip = ZipFile(filename, 'r')
        filelist = [entry.replace('\\', '/') for entry in zip.namelist()]
        currentdir = os.getcwd()
        os.chdir(directorypath)         # so that file access is relative

        delfiles = [entry.replace('\\', '/') for entry in self.delfiles]          # so that paths are always the same 
        deldirs = [entry.replace('\\', '/') for entry in self.deldirs]
        newfiles = [entry.replace('\\', '/') for entry in (self.modfiles + self.newfiles)]
        newdirs = [entry.replace('\\', '/') for entry in self.newdirs]
                       
        while delfiles:
            thisfile = delfiles.pop()
            try:
                os.remove(thisfile) 
            except:
                error = 'Error removing file %s.' % thisfile
                if warnings == 1:
                    self.warnings.append(error)
                elif warnings == 2:
                    print error
                    raise

        while deldirs:
            thisdir = deldirs.pop()
            try:
                rmtree(thisdir) 
            except:
                error = 'Error removing directory %s.' % thisdir
                if warnings == 1:
                    self.warnings.append(error)
                elif warnings == 2:
                    print error
                    raise
                    
        for name in filelist:
            if name in newfiles:
                newfiles.remove(name)
            if not name.endswith('/'):
                try:
                    thisfile = zip.read(name)
                except:
                    error = 'Error extracting file %s.' % name
                    if warnings == 1:
                        self.warnings.append(error)
                        continue
                    elif warnings == 2:
                        print error
                        raise               
                try:
                    pathdir = dirname(name)
                    if pathdir and not exists(pathdir): os.makedirs(pathdir)
                    writeb(name, thisfile)
                except:
                    error = 'Error creating file %s.' % name
                    if warnings == 1:
                        self.warnings.append(error)
                    elif warnings == 2:
                        print error
                        raise
            else:
                try:
                    if name and not isdir(name): os.makedirs(name)
                except:
                    error = 'Error creating directory %s.' % name
                    if warnings == 1:
                        self.warnings.append(error)
                    elif warnings == 2:
                        print error
                        raise
                    
        if newfiles and warnings:
            error = 'Files left in filelist, not found in the change object/script - %s' % ', '.join(newfiles)
            if warnings == 2:
                raise Exception(': ' + error)
            elif warnings == 1:
                self.warnings.append(error)

        for name in newdirs:
            if not exists(name):
                error = 'Directory Warning - directory %s may not have been created.' % name
                if warnings == 1:
                    self.warnings.append(error)
                elif warnings == 2:
                    raise Exception(': ' + error) 
                os.makedirs(name)

        zip.close()      
        os.chdir(currentdir)


#####################################################
    # functions to do with tags

def bquote(element):
    """basic quote - Put quotes around an element."""
    if not isinstance(element, (str, unicode)): element = str(element)
    return '\"' + element.replace('\"', '&mjf-quot;') + '\"'

def maketag(tag, value=None):
    """Create a tag from the tag name and it's value."""
    if not isinstance(value, (str, unicode, type(None))): value = str(value)
    tag = tag.upper()
    if value == None: return '<'+tag+'>'
    return '<'+tag+' = '+bquote(value)+'>'

def handletag(line):
    """Extract a tag, value from a tagged line.
    If there is no value then value is None
    line should already be stripped.
    Returned tags are lowercase.
    FORMAT : <tag = "value"> *or* <tag> *or* </tag>
    If the line is invalid, the returned tag is None (and value is None)
    """
    if not line.startswith('<') or not line.endswith('>'):
        return None, None
    line = line[1:-1].strip()       # remove the end and finish tag marker
    tag = ''
    if line[0] == '/':
        tag = '/'
        line = line[1:].lstrip()
    index = -1
    while index < len(line)-1:
        index += 1
        char = line[index]
        if char == ' ': break   # if we hit a space then we have reached the start of the value
        if char == '=': break   # if we hit = we have also reached the start of the value
        tag += char
    else:
        return tag.lower(), None
    line = line[index:].lstrip()
    if line[0] != '=': return None, None        # something other than a value inside the tag
    line=line[1:].lstrip()
    if line[0] == '\"':                         # if value is quoted, remove value      - only accepts \" for quoting
        line = line[1:]
        if line[-1] != '\"':  return None, None        # badly quoted value
        line = line[:-1]
    return tag.lower(), line    

#######################################################################
    # a pre 2.3 walkfiles function - adapted from the path module by Jason Orendorff

def walkfiles(thisdir):
    """ walkfiles(D) -> iterator over files in D, recursively.
    """
    for child in os.listdir(thisdir):
        thischild = join(thisdir, child)
        if isfile(thischild):
            yield thischild
        elif isdir(thischild):
            for f in walkfiles(thischild):
                yield f
            
def walkemptydirs(thisdir):
    """This function is used to find empty directories - which will be missed by the walkfiles function."""
    if not os.listdir(thisdir):       # if the directory is empty.. then yield it
        yield thisdir   
    for child in os.listdir(thisdir):
        thischild = join(thisdir, child)
        if isdir(thischild):
            for emptydir in walkemptydirs(thischild):
                yield emptydir


#######################################################################
    
def FSDMdetails(filename):
    """Given a filename it returns False if the file doesn't look like an FSDM file (or doesn't exist).
    If it does look like an FSDM file it returns a tuple of the FSDM root attributes :
    (archivename, thepath, thesize, archivedate, numfiles, numdirs)
    It assumes the FSDM is a normal, valid structure.
    thesize, numfiles, numdirs are integers - the other values are strings.
    """
    if not isfile(filename): return False
    try:
        indir = readl(filename, True)
    except:
        return False
    index = -1
    donestart = 0
    archivename = thepath = thesize = archivedate = numfiles = numdirs = -1
    while index < len(indir)-1:          # indir is the FSDM structure we are parsing - it will either be just a directory structure or a whole FSDM struct
        index += 1
        line = indir[index]
        tag, value = handletag(line)
        if tag ==  'fsdm' and not donestart:
            donestart = 1
            continue
        elif not donestart:
            continue
        elif not tag:               # invalid tag or line - return False ?
            continue
        elif tag=='directory' or tag=='file':          # when we get to the first file or directory tag we've finished
            break
        elif tag == 'aname':
            archivename = value
            continue
        elif tag == 'size':
            thesize = int(value)
            continue
        elif tag == 'numfiles':
            numfiles = int(value)
            continue
        elif tag == 'numdirs':
            numdirs = int(value)
            continue
        elif tag == 'abspath':
            thepath = value
            continue
        elif tag == 'date':
            archivedate = value
            continue
    returnval = (archivename, thepath, thesize, archivedate, numfiles, numdirs)
    for entry in returnval:
        if entry == -1: return False
    return returnval

#######################################################################
    # a few functions for dealing with files
def readb(filename):
    """This returns a file as a binary string."""
    fileh = file(filename, 'rb')
    outstuff = fileh.read()
    fileh.close()
    return outstuff

def readl(filename, stripopt=False):
    """Return a filename as a list of lines."""
    fileh = file(filename, 'r')
    outstuff = fileh.readlines()
    fileh.close()
    if stripopt: return [entry.strip() for entry in outstuff]
    return outstuff
    
def writeb(filename, instuff):
    """This writes a file as a binary string."""
    fileh = file(filename, 'wb')
    fileh.write(instuff)
    fileh.close()

def writel(filename, inlines):
    """This writes a file as a list of lines.
    Terminates each line in '\n' if necessary."""
    fileh = file(filename, 'w')
    outlines = []
    for line in inlines:
        if not line.endswith('\n'):
            line = line + '\n'
        fileh.write(line)    
    fileh.close()


########################################################################
    # test function

def test(testdir = 'F:\Python Projects\Tools'):
    """As it stands this code profiles a directory (testdir),
    it writes out the profile as an FSDM structure and generates a new DirObj by reading it in again (using ReadFSDM)
    It then asks you to make a change to a directory, and DirObj and ChangeObject profile the change.
    You can then undo the change and watch the ChangeObject put it back !!
    The commented out code displays some of the details from the structures it creates -
    it compares the DirObj from the first profile to the one it makes when it reads the FSDM structure back in.
    """

    time1 = time.time()
    print 'Working in %s' % testdir
    testobj = DirObj(testdir, name='Python Projects')
    time2 = time.time()
    outfile = testobj.write()
    time3 = time.time()
    testobj2 = ReadFSDM(outfile)
    time4 = time.time()

    print 'Read time from disk :', time2 - time1
    print 'Write method :', time3 - time2
    print 'Read from FSDM structure :', time4 - time3
    print 'Total Size of file structure : ', testobj.size, ' bytes'
    print 'Size read from saved script : ', testobj2.size, ' bytes'

##    print testobj.files
##    size = 0
##    for line in outfile:
##        size += len(line)
##        print line 
##    print size

##    outfile2 = testobj2.write()
##    print testobj2.dirs
##    size = 0
##    for line in outfile2:
##        size += len(line)
###        print line 
##    print size
### so we have two stuctures - a testobj created from a directory structure
### and a testobj 2 built back from the markup that testobj created
##    print len(outfile)-len(outfile2)
##    
##    print 'Files:'
##    for afile in  testobj.files:
##        print afile
####        for line in afile.write():
####            print line
##    print
##    print testobj.dirs.keys()
##    print testobj2.dirs.keys()
##    print 'Directories:'
##    for adir in testobj2.dirs:
##        print
##        print testobj.dirs[adir].name
##        for anotherdir in testobj.dirs[adir].dirs:
##            print anotherdir
##            print testobj.dirs[adir].dirs[anotherdir].path, 'name', testobj.dirs[adir].dirs[anotherdir].name
##            print testobj2.dirs[adir].dirs[anotherdir].path, 'name', testobj2.dirs[adir].dirs[anotherdir].name
##            print

    dummy = raw_input('Make Change...')
    if dummy.strip().lower() == 'z': return
    testobj3 = DirObj(testdir, name='Python Projects')
    time5 = time.time()
    change = ChangeObject()
    change.compare(testobj2, testobj3)
    time6 = time.time()
    print 'Creating change object and comparing the structures : ', time6 - time5, ' seconds....'
    if change.write():
        for line in change.write():
            print line
        change.build(testdir, 'test.zip')

        print 'Making changes from Zip'
        dummy = raw_input('Undo Change...')
        time7 = time.time()
        change.rebuild(testdir, 'test.zip')
        time8 = time.time()
        print 'Rebuilding complete : ', time8 - time7, ' seconds....'
    else:
        print 'No Changes Made'

def test2(testdir="G:\\"):
    """A very simple test that just creates a structure, reads it and compares it to itself."""
    testobj = DirObj(testdir)
    test = testobj.write()
    testobj2 = ReadFSDM(test)
    print testobj2.dirs
    for entry in testobj2.dirs:
        print testobj2.dirs[entry].path

    testobj.creationtime += 1
    change = ChangeObject()
    change.compare(testobj2, testobj)
    return (change, testobj, testobj2)


########################################################################

# the following code is used by filestruct when it's run from the command line

def displayhelp():
    helpstring = """FILESTRUCT

filestruct is a command line tool that will help you keep two directories in different locations in sync. It does that by recording changes in the file structure and recording the differences. From the zip file that it saves it is able to then make all the changes that have occurred in the remote location. If you then change any of the files filestruct can produce a single file of all the changes you have made.

The basic command line syntax is very simple :

filestruct.py action args

filestruct can perform three different actions - create, compare and update.
Each action has a slightly different command line synatx - but they are very simple and logical.
Each action relates to a different stage of the process.

filestruct.py create pathname outfilename
filestruct.py compare pathname infilename outfilename
filestruct.py update pathname infilename

create
This will examine the contents of a file structure and save a representation of it using FSDM (File Structure Description Markup).
Basically you give it a pathname (directory) to examine and it saves out a text file describing the structure.

filestruct.py create pathname outfilename
pathname is the path to the file structure you want it to examine (a directory).
outfilename is the filename to save the description file out as.

compare
This takes an FSDM file created by the create method and compares it with the contents of a pathname.
It will generate a zip file that contains a description of the changes and all the files needed to make the changes.
Inside is a text file describing the changes and a zip pf all the files it needs.

filestruct.py compare pathname infilename outfilename
pathname is the path to the file structure you want it to examine (a directory).
infilename is a previously generated FSDM file.
outfilename is the name to save the zipfile it produces.

update
This takes a zip file produced by compare and makes all the changes in it.
WARNING - this program will delete files and directories if that's what it's told to do.

filestruct.py update pathname infilename
pathname is the name of the file structure to update.
infilename is the zip file to do it from.


This utility is only really useful if you are maintaining two *identical* file structures on seperate computers. I have two remote folders that I regularly transfer from work to home (almost daily) - a research folder and a python projects folder. This program does this job very well - and has the library code to do lot's of other jobs - but I can imagine the command line tool causing a lot of damage if misused. It especially shouldn't be used if you are changing both directories *seperately* and then trying to sync them !!

There is an additional command line argument -w or --warnings.
This is a very simple way of determining the way that filestruct handles warnings.
It should be a value from 0-2
0 - means ignore all warnings, don't display them
1 - (the default) print warnings, but continue
2 - stop if a serious error occurs.

If you want to change the warning level from the default (1), you should set it before the other arguments.
e.g.
filestruct.py -w2 compare pathname infilename outfilename
filestruct.py --warnings 2 compare pathname infilename outfilename

Just running filestruct with no arguments should display all of this, so should filestruct -h, filestruct --help, filestruct ?

Speed of indexing large structures is very dependent on disk-access times.
Indexing an 873MByte directory, with 16647 files in 668 folders took me 31.25 seconds.
With psyco enabled it only took 20.14 seconds.
A repeat of the read (most of the disk information cached) took 7 seconds.
Parsing the FSDM structure representing the same File Structure took only 2.125 seconds.

filestruct.py also contains a useful library of python objects for reading filestructures, creating FSDM structures, reading and comparing them etc.
Fuzzyman - http://www.voidspace.org.uk/atlantibots/pythonutils.html
"""
    helplines = helpstring.splitlines()
    while helplines:
        theselines = helplines[:15]
        helplines = helplines[15:]
        for line in theselines:
            print line
        try:
            dummy = raw_input('Enter to continue..... X to exit....')
            if dummy.lower().strip() == 'x': break
        except:
            break
    sys.exit(0)

def create(commline, warning):
    """This implements the create action from the command line."""
    if len(commline) != 2:
        print 'create action requires only three arguments - '
        print """filestruct.py create pathname outfilename
pathname is the path to the file structure you want it to examine (a directory).
outfilename is the filename to save the description file out as."""
        sys.exit(1)
    pathname = commline[0]
    outfilename = commline[1]
    if not isdir(pathname):
        print 'Pathname given %s isn\'t a valid directory.' % pathname
        sys.exit(1)
    dirobj = DirObj(pathname, name='Produced by filestruct.py')
    outfile = dirobj.write()
    try:
        writel(outfilename, outfile)
    except:
        print 'Error whilst writing out the file %s' % outfilename
        raise

def compare(commline, warning):
    """This implements the compare action from the command line."""
    if len(commline) != 3:
        print 'compare action requires four arguments - '
        print """filestruct.py compare pathname infilename outfilename
pathname is the path to the file structure you want it to examine (a directory).
infilename is a previously generated FSDM file.
outfilename is the name to save the zipfile it produces."""
        sys.exit(1)
    pathname = commline[0]
    infilename = commline[1]
    outfilename = commline[2]
    if not isdir(pathname):
        print 'Pathname given %s isn\'t a valid directory.' % pathname
        sys.exit(1)
    if not isfile(infilename):
        print 'Infilename given %s isn\'t a valid file.' % infilename
        sys.exit(1)
    dirobj1 = ReadFSDM(readl(infilename))
    dirobj2 = DirObj(pathname, name='Produced by filestruct.py')
    changeobj = ChangeObject()
    changeobj.compare(dirobj1, dirobj2)
    test = changeobj.build(pathname, outfilename, warning)
    if not test:
        print 'No changes made, zipfile not created.'
        return
    if changeobj.warnings:
        for line in changeobj.warnings: print line

def update(commline, warning):
    """This implements the update action from the command line."""
    if len(commline) != 2:
        print 'update action requires only three arguments - '
        print """filestruct.py update pathname infilename
pathname is the name of the file structure to update.
infilename is the zip file to do it from."""
        sys.exit(1)
    pathname = commline[0]
    infilename = commline[1]
    if not isdir(pathname):
        print 'Pathname given %s isn\'t a valid directory.' % pathname
        sys.exit(1)
    if not isfile(infilename):
        print 'Infilename given %s isn\'t a valid file.' % infilename
        sys.exit(1)
    changeobj = ChangeObject()
    changeobj.rebuild(pathname, infilename, warning)
    if changeobj.warnings:
        for line in changeobj.warnings: print line


def main():
    """The command line tool part.
    This bit sorts the arguments out and hands over to the relevant function."""
    if len(sys.argv) > 1:
        commline = sys.argv[1:]
    else:
        displayhelp()
    helps = ['-h', 'help', '?', '--help']
    if commline[0].lower() in helps: displayhelp()
    if len(commline) < 3:
        print 'filestruct needs at least three arguments when run from the command line. See :'
        print 'filestruct.py ?'
        sys.exit(1)

    warning = 1
    if commline[0].lower().startswith('-w') or commline[0].lower().startswith('--warning'):
        if commline[0].lower().startswith('-w'):
            wline = commline[0][2:]
            sbit = '-w'
        else:
            sbit = '--warning'
            wline = commline[0][9:]
        if wline and not wline.isdigit():
            print 'Unreconginsed option %s' % commline[0]
            print 'filestruct.py ?'
            print 'For full command line syntax.'
            sys.exit(1)
        elif wline and int(wline) <= 2:
            warning = int(wline)
            commline = commline[1:]
        else:
            if wline or not commline[1].isdigit() or int(commline[1]) > 2:
                print sbit + ' option requires a digit between 0 and 2 immediately following.'
                sys.exit(1)
            warning = int(commline[1])
            commline = commline[2:]
# this has dealt with and removed the --warning argument if it is present
# we *still* need at least three command line arguments
    if len(commline) < 3:
        print 'filestruct needs at least three arguments when run from the command line. See :'
        print 'filestruct.py ?'
        sys.exit(1)
    action = commline.pop(0)
    actiondict = { 'create' : create, 'compare' : compare, 'update' : update }
    if action.lower() not in ['create', 'compare', 'update']:
        print "action argument must be 'create', 'compare' or 'update'. See :"
        print 'filestruct.py ?'
        sys.exit(1)
    else:
        actiondict[action](commline, warning)           # who says python doesn't have a switch command !

if __name__ == '__main__':

    main()
#    test2()
    
########################################################################


"""
COMMAND LINE TOOL

try running - filestruct ?
or see the helpstring in the displayhelp function.



TODO/ISSUES
Additonal library support to make 'Simple Version Control'.
Invalid structures ought to cause it to throw a format exception when parsing. (At the moment it just creates a bad structure..)
Add tar support.
Use temporary directories properly. At the moment it uses the current directory - which will cause errors if run from a CD for example. (or see if we can write directly into a zip file without using temporary directories).
Use the caseless module so we have case insensitive file checking. (only applies to windows anyway - maybe check what system we're on ?).
ASCII write mode - for unpacking text type files across platforms.
Test zip files before attempting to read them.
Doesn't check that files in the archive are ones from the FSCO script - does check for extra ones though. (because files from the new directory list don't appear in the new file list anyway - and would throw false errors). 
Create a proper Exception. (Using the generic 'Exception' in a couple of places).
Save file flags in the file attributes. (So need to write code to change attributes and add a new section to FSCO !)
Won't work across different systems with a different 'epoch' (where the same date/time is represented using a different number for times like mtime and ctime of a file).
Support for updating by FTP (checking from FTP ? - probably not !!)


CHANGELOG
Version 1.0.4       6th July 2004
Added the ZIP_DEFLATED constant. This module wasn't compressing the zip files it created !! 

Version 1.0.3       15th June 2004
Will now use the ordereddict module if it is available.
This creates all dictionaries using the oDict object, meaning that markup structures will always be created consistently.

Version 1.0.2       13th June 2004
Added the numfiles and numdirs attributes to DirObj and ReadFSDM objects.
If no changes have occurred (or the changeobject is uninitialized) change.write() returns None. (unless the ChangeObject.debug attribute is set to 1)

Version 1.0.1       9th June 2004
Handles different '\' and '/' in paths. (Was incorrectly reporting files not found in the zipfile).
Will now write and unpack empty directories.
Added the FSDMdetails function.

Version 1.0.0       7th June 2004
First version, will create, read and compare FSDM files.
Can also implement changes. Command line version is basic but works 'as advertised'...
"""
